var snowbox = (function (exports, axios, normalizr, react, reselect, redux, produce, _) {
  'use strict';

  function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

  var axios__default = /*#__PURE__*/_interopDefaultLegacy(axios);
  var produce__default = /*#__PURE__*/_interopDefaultLegacy(produce);
  var ___default = /*#__PURE__*/_interopDefaultLegacy(_);

  const NAMESPACE = 'snowbox';
  const actions = {
    UPSERT: `${NAMESPACE}/UPSERT`,
    REMOVE: `${NAMESPACE}/REMOVE`,
    FIND: `${NAMESPACE}/FIND`,
    FETCH: `${NAMESPACE}/FETCH`,
    CLEAR: `${NAMESPACE}/CLEAR`,
    NO_FETCH: `${NAMESPACE}/NO_FETCH`
  };
  const statuses = {
    PENDING: 'pending',
    SUCCEEDED: 'succeeded',
    FAILED: 'failed'
  };
  const contentTypes = {
    JSON: `${NAMESPACE}/JSON`,
    FORM_DATA: `${NAMESPACE}/FORM_DATA`
  };

  const SUCCESS_EXTENSION = '_SUCCESS';
  const FAILURE_EXTENSION = '_FAILURE';
  const isSuccess = type => {
    const regExp = new RegExp(`${SUCCESS_EXTENSION}$`);
    return regExp.test(type);
  };
  const isFailure = type => {
    const regExp = new RegExp(`${FAILURE_EXTENSION}$`);
    return regExp.test(type);
  };
  const success = str => `${str}${SUCCESS_EXTENSION}`;
  const failure = str => `${str}${FAILURE_EXTENSION}`;
  const createAction = type => (payload = {}) => ({
    type,
    ...payload
  });
  const createSuccess = type => entity => (payload, entities, result, meta, date) => ({
    type: success(type),
    entity,
    payload,
    entities,
    result,
    meta,
    date
  });
  const createFailure = type => entity => (payload, error, statusCode) => ({
    type: failure(type),
    entity,
    payload,
    error,
    statusCode
  });
  const request = type => entity => (payload, meta = {}) => ({
    type,
    entity,
    payload,
    meta,
    success: createSuccess(type)(entity),
    failure: createFailure(type)(entity)
  });
  const upsert = request(actions.UPSERT);
  const remove = request(actions.REMOVE);
  const find = request(actions.FIND);
  const fetch = request(actions.FETCH);
  const clearAll = createAction(actions.CLEAR);
  const noFetch = createAction(actions.NO_FETCH);

  const getContentTypeHeader = contentType => {
    switch (contentType) {
      case contentTypes.JSON:
        return {
          'Content-type': 'application/json; charset=utf-8'
        };

      case contentTypes.FORM_DATA:
        return {
          'Content-type': 'multipart/form-data'
        };

      default:
        throw new Error(`[Snowbox API] Invalid content type "${contentType}"`);
    }
  };

  const getAuthHeader = async (tokenHeader, getAuthToken) => {
    if (tokenHeader && getAuthToken) {
      return {
        [tokenHeader]: await getAuthToken()
      };
    }

    return {};
  };

  const validateBaseUrl = providedOptions => {
    if (typeof providedOptions.baseUrl !== 'string' || providedOptions.baseUrl === '') {
      throw new Error('[Snowbox API] Base API url must be defined');
    }
  };

  const validateAuthOptions = providedOptions => {
    const isTokenHeaderValid = typeof providedOptions.tokenHeader === 'string' && providedOptions.tokenHeader.length > 0;
    const isGetAuthTokenValid = typeof providedOptions.getAuthToken === 'function';

    if (isTokenHeaderValid !== isGetAuthTokenValid) {
      throw new Error('[Snowbox API] Both tokenHeader and getAuthToken must be provided');
    }
  };

  const setupOptions = providedOptions => ({
    baseUrl: null,
    tokenHeader: null,
    getAuthToken: null,
    ...providedOptions
  });

  const api = (providedOptions = {}) => {
    validateBaseUrl(providedOptions);
    validateAuthOptions(providedOptions);
    const options = setupOptions(providedOptions);
    const axiosInstance = axios__default['default'].create({
      baseURL: options.baseUrl
    });

    const get = (path, params) => request('get', path, params);

    const post = (path, data = {}, params, contentType) => request('post', path, params, data, contentType);

    const put = (path, data = {}, params, contentType) => request('put', path, params, data, contentType);

    const patch = (path, data = {}, params, contentType) => request('patch', path, params, data, contentType);

    const remove = path => request('delete', path);

    const request = async (method, path, params, data, contentType = contentTypes.JSON) => {
      const authHeader = await getAuthHeader(options.tokenHeader, options.getAuthToken);
      const requestConfig = {
        params,
        data,
        headers: { ...getContentTypeHeader(contentType),
          ...authHeader
        }
      };

      if (contentType === contentTypes.FORM_DATA) {
        requestConfig.transformRequest = [data => {
          if (!data || typeof data !== 'object') {
            throw new Error('[Snowbox API] data must be object');
          }

          const body = new FormData();
          Object.keys(data).forEach(field => body.append(field, data[field]));
          return body;
        }];
      }

      if (method === 'get' || method === 'delete') {
        return axiosInstance[method](path, requestConfig);
      }

      return axiosInstance[method](path, data, requestConfig);
    };

    return {
      get,
      post,
      put,
      patch,
      remove,
      request
    };
  };

  const entity = (key, provider, relations = {}, options = {}) => {
    var _provider$options;

    // Needed by normalizr.
    options.idAttribute = options.idField || (provider === null || provider === void 0 ? void 0 : (_provider$options = provider.options) === null || _provider$options === void 0 ? void 0 : _provider$options.idField) || 'id';
    const newEntity = new normalizr.schema.Entity(key, relations, options);

    if (provider) {
      if (typeof provider !== 'object') {
        throw new Error('[Snowbox] Invalid provider');
      }

      newEntity.provider = provider;
    }

    newEntity.idField = options.idAttribute;

    if (typeof options.staleTimeout === 'number') {
      newEntity.staleTimeout = options.staleTimeout;
    }

    if (options.singleton === true) {
      newEntity.singleton = true;
    }

    return newEntity;
  };

  const baseKey = '#';
  const buildKey = (filters = {}) => {
    if (!filters || typeof filters != 'object' || Array.isArray(filters)) {
      return baseKey;
    }

    return Object.keys(filters).sort().reduce((key, filter) => {
      key += `${filter}[${filters[filter]}]`;
      return key;
    }, baseKey);
  };

  const selectSnowbox = state => state.snowbox;

  const selectCollections = reselect.createSelector(selectSnowbox, snowbox => snowbox.collections);
  const selectEntities = reselect.createSelector(selectSnowbox, state => state.entities);
  const selectSingletons = reselect.createSelector(selectSnowbox, state => state.singletons);

  const selectEntityCollection = entity => (state, filters) => {
    var _state$snowbox, _state$snowbox$collec;

    const collections = state === null || state === void 0 ? void 0 : (_state$snowbox = state.snowbox) === null || _state$snowbox === void 0 ? void 0 : (_state$snowbox$collec = _state$snowbox.collections) === null || _state$snowbox$collec === void 0 ? void 0 : _state$snowbox$collec[entity.key];
    return collections === null || collections === void 0 ? void 0 : collections[buildKey(filters)];
  };

  const applyHydration = (entity, levels = 0) => entities => data => {
    if (!data || !levels) {
      return data;
    }

    if (Array.isArray(data)) {
      return data.map(item => applyHydration(entity, levels)(entities)(item));
    }

    const newData = { ...data
    };

    for (const key in entity.schema) {
      const childEntity = Array.isArray(entity.schema[key]) ? entity.schema[key][0] : entity.schema[key];
      newData[key] = applyHydration(childEntity, levels - 1)(entities)(Array.isArray(data[key]) ? data[key].map(id => entities[childEntity.key][id]) : entities[childEntity.key][data[key]]);
    }

    return newData;
  };

  const selectOne = (entity, hydrationLevels = 0, idAttribute) => {
    if (entity.singleton) {
      return reselect.createSelector(selectSingletons, singletons => singletons[entity.key]);
    }

    const hydrate = applyHydration(entity, hydrationLevels);
    return reselect.createSelector(selectEntities, (state, props) => props[idAttribute || entity.idField], (entities, id) => entities[entity.key] ? hydrate(entities)(entities[entity.key][id]) : null);
  };
  const emptyArray = [];
  const selectCollection = (entity, hydrationLevels = 0) => {
    const hydrate = applyHydration(entity, hydrationLevels);
    return reselect.createSelector(selectEntities, selectEntityCollection(entity), (allEntities, collection) => {
      if (!allEntities[entity.key] || !collection || !collection.result) {
        return emptyArray;
      }

      const selectedEntities = collection.result.map(id => allEntities[entity.key][id]).filter(obj => obj);
      return hydrate(allEntities)(selectedEntities);
    });
  };
  const selectAll = (entity, hydrationLevels = 0) => {
    const hydrate = applyHydration(entity, hydrationLevels);
    return reselect.createSelector(selectEntities, allEntities => hydrate(allEntities)(Object.values(allEntities[entity.key] || {})));
  };
  const selectMeta = entity => reselect.createSelector(selectEntityCollection(entity), collection => collection === null || collection === void 0 ? void 0 : collection.meta);

  const methods = {
    [actions.UPSERT]: 'upsert',
    [actions.REMOVE]: 'remove',
    [actions.FIND]: 'find',
    [actions.FETCH]: 'fetch'
  };

  const getEntityFromAction = action => Array.isArray(action.entity) ? action.entity[0] : action.entity;

  const getRequestMethodFromAction = action => methods[action.type];

  const selectExistingData = (store, action) => {
    var _state$entity$key;

    const entity = getEntityFromAction(action);

    if (Array.isArray(action.entity)) {
      var _action$meta, _selectCollections;

      const key = buildKey(action.payload);
      const state = action !== null && action !== void 0 && (_action$meta = action.meta) !== null && _action$meta !== void 0 && _action$meta.isListHook ? store.getState() : (_selectCollections = selectCollections(store.getState())) === null || _selectCollections === void 0 ? void 0 : _selectCollections[entity.key];
      return state === null || state === void 0 ? void 0 : state[key];
    }

    if (entity.singleton) {
      const state = selectSingletons(store.getState());
      return state === null || state === void 0 ? void 0 : state[entity.key];
    }

    const state = selectEntities(store.getState());
    const id = action.payload[entity.idField];
    return state === null || state === void 0 ? void 0 : (_state$entity$key = state[entity.key]) === null || _state$entity$key === void 0 ? void 0 : _state$entity$key[id];
  };
  const dataAlreadyExists = (store, action) => {
    var _action$meta2;

    const method = getRequestMethodFromAction(action);
    const entity = getEntityFromAction(action);

    if (!['find', 'fetch'].includes(method)) {
      return false;
    }

    if (!entity.staleTimeout) {
      return false;
    }

    if (action !== null && action !== void 0 && (_action$meta2 = action.meta) !== null && _action$meta2 !== void 0 && _action$meta2.refresh) {
      return false;
    }

    let state = selectExistingData(store, action);

    if (!(state !== null && state !== void 0 && state.__updatedAt)) {
      return false;
    }

    return Date.now() - state.__updatedAt < entity.staleTimeout;
  };

  const isHandledAction = action => [actions.UPSERT, actions.REMOVE, actions.FIND, actions.FETCH].includes(action.type);

  const requestData = action => {
    const method = getRequestMethodFromAction(action);
    const entity = getEntityFromAction(action);
    return entity.provider[method](action.payload);
  };

  const getSuccessActionParams = (action, response) => {
    var _action$meta3;

    const method = getRequestMethodFromAction(action);

    if (method === 'remove') {
      return [action.payload, undefined, response.original];
    }

    const entity = getEntityFromAction(action);

    if (entity.singleton) {
      return [action.payload, undefined, response.data, undefined, Date.now()];
    }

    if (action !== null && action !== void 0 && (_action$meta3 = action.meta) !== null && _action$meta3 !== void 0 && _action$meta3.isListHook) {
      return [action.payload, undefined, response.data, response.meta, Date.now()];
    }

    const {
      entities,
      result
    } = normalizr.normalize(response.data, action.entity);
    return [action.payload, entities, result, response.meta, Date.now()];
  };

  const snowboxMiddleware = store => next => async action => {
    var _action$meta4;

    if (!isHandledAction(action)) {
      return next(action);
    }

    if (dataAlreadyExists(store, action)) {
      return next(noFetch());
    }

    if (!(action !== null && action !== void 0 && (_action$meta4 = action.meta) !== null && _action$meta4 !== void 0 && _action$meta4.isListHook)) {
      next(action);
    }

    try {
      const response = await requestData(action);
      const actionParams = getSuccessActionParams(action, response);
      return next(action.success(...actionParams));
    } catch (error) {
      return next(action.failure(action.payload, error, error.status));
    }
  };

  const snowboxActions = [actions.FETCH, success(actions.FETCH), failure(actions.FETCH), actions.FIND, success(actions.FIND), failure(actions.FIND), actions.UPSERT, success(actions.UPSERT), failure(actions.UPSERT), actions.REMOVE, success(actions.REMOVE), failure(actions.REMOVE)];

  const entitiesReducer = (state = {}, action) => {
    var _action$entity;

    if (!snowboxActions.includes(action.type)) {
      return state;
    }

    if (action !== null && action !== void 0 && (_action$entity = action.entity) !== null && _action$entity !== void 0 && _action$entity.singleton) {
      return state;
    }

    if (action.type === success(actions.REMOVE)) {
      const id = typeof action.payload === 'number' ? action.payload : action.payload[action.entity.idField];
      return produce__default['default'](state, draftState => {
        delete draftState[action.entity.key][id];
      });
    }

    if (!action.entities) {
      return state;
    }

    return produce__default['default'](state, draftState => {
      for (const entity in action.entities) {
        draftState[entity] = draftState[entity] || {};

        for (const id in action.entities[entity]) {
          draftState[entity][id] = { ...draftState[entity][id],
            ...action.entities[entity][id],
            __updatedAt: action.date
          };
        }
      }
    });
  };

  const fetchEntityCollectionsReducer = (state = {}, action) => {
    switch (action.type) {
      case actions.FETCH:
        return {
          status: statuses.PENDING
        };

      case success(actions.FETCH):
        return {
          status: statuses.SUCCEEDED,
          result: action.result,
          meta: action.meta,
          __updatedAt: action.date
        };

      case failure(actions.FETCH):
        return {
          status: statuses.FAILED,
          error: action.error
        };

      default:
        return state;
    }
  };

  const entityCollectionsReducer = (state = {}, action) => {
    const key = buildKey(action.payload);
    return produce__default['default'](state, draftState => {
      draftState[key] = fetchEntityCollectionsReducer(state[key], action);
    });
  };

  const collectionsReducer = (state = {}, action) => {
    const allowedActions = [actions.FETCH, success(actions.FETCH), failure(actions.FETCH)];

    if (!allowedActions.includes(action.type)) {
      return state;
    }

    if (!Array.isArray(action.entity)) {
      return state;
    }

    const entity = action.entity[0];
    return produce__default['default'](state, draftState => {
      draftState[entity.key] = entityCollectionsReducer(state[entity.key], action);
    });
  };

  const singletonsReducer = (state = {}, action) => {
    var _action$entity2;

    if (!snowboxActions.includes(action.type)) {
      return state;
    }

    if (!(action !== null && action !== void 0 && (_action$entity2 = action.entity) !== null && _action$entity2 !== void 0 && _action$entity2.singleton)) {
      return state;
    }

    if (action.type === success(actions.REMOVE)) {
      return produce__default['default'](state, draftState => {
        delete draftState[action.entity.key];
      });
    }

    return produce__default['default'](state, draftState => {
      draftState[action.entity.key] = { ...action.result,
        __updatedAt: action.date
      };
    });
  };

  const rootReducer = redux.combineReducers({
    entities: entitiesReducer,
    collections: collectionsReducer,
    singletons: singletonsReducer
  });
  const snowboxReducer = (state, action) => {
    if (action.type === actions.CLEAR) {
      return rootReducer(undefined, action);
    }

    return rootReducer(state, action);
  };

  const fetchList = (middleware, dispatch, entity, filters, refresh) => {
    middleware(dispatch)(fetch([entity])(filters, {
      isListHook: true,
      refresh
    }));
  };

  const useList = (entity, initialFilters = {}) => {
    var _state$key, _state$key2, _state$key3, _state$key4, _state$prevKey, _state$prevKey2;

    const [filters, setCurrFilters] = react.useState(initialFilters);
    const [prevFilters, setPrevFilters] = react.useState(null);
    const [state, dispatch] = react.useReducer(entityCollectionsReducer, {});
    const middleware = react.useMemo(() => snowboxMiddleware({
      getState: () => state
    }), [state]);
    react.useEffect(() => {
      fetchList(middleware, dispatch, entity, filters);
    }, []);

    const write = (payload, actionCreator) => new Promise((resolve, reject) => {
      const nextOnWrite = action => {
        if (isSuccess(action.type)) {
          resolve(action.result);
          fetchList(middleware, dispatch, entity, filters, true);
        }

        if (isFailure(action.type)) {
          reject(action.error);
        }
      };

      middleware(nextOnWrite)(actionCreator(entity)(payload, {
        isListHook: true
      }));
    });

    const setFilters = newFilters => {
      setPrevFilters(filters);
      setCurrFilters(newFilters);
      fetchList(middleware, dispatch, entity, newFilters);
    };

    const upsert$1 = payload => write(payload, upsert);

    const remove$1 = payload => write(payload, remove);

    const key = buildKey(filters);
    const prevKey = buildKey(prevFilters);
    return {
      items: ((_state$key = state[key]) === null || _state$key === void 0 ? void 0 : _state$key.result) || [],
      meta: ((_state$key2 = state[key]) === null || _state$key2 === void 0 ? void 0 : _state$key2.meta) || {},
      filters,
      status: ((_state$key3 = state[key]) === null || _state$key3 === void 0 ? void 0 : _state$key3.status) || statuses.PENDING,
      error: (_state$key4 = state[key]) === null || _state$key4 === void 0 ? void 0 : _state$key4.error,
      prevFilters,
      prevItems: (_state$prevKey = state[prevKey]) === null || _state$prevKey === void 0 ? void 0 : _state$prevKey.result,
      prevMeta: (_state$prevKey2 = state[prevKey]) === null || _state$prevKey2 === void 0 ? void 0 : _state$prevKey2.meta,
      setFilters,
      upsert: upsert$1,
      remove: remove$1
    };
  };

  class Response {
    constructor(response, options, isFetch) {
      this._response = response;
      this._isFetch = isFetch;
      this._entityPath = options.entityPath;
      this._entitiesPath = options.entitiesPath;
      this._entitiesFieldName = options.entitiesFieldName;
      this._hasMeta = options.hasMeta;
      this._metaPath = options.metaPath;
      this._metaFieldName = options.metaFieldName;
    }

    get data() {
      return this._isFetch ? ___default['default'].get(this._response, this._entitiesPath, this._response) : ___default['default'].get(this._response, this._entityPath, this._response);
    }

    get meta() {
      if (!this._isFetch || !this._hasMeta) {
        return undefined;
      }

      const responseWithoutEntities = ___default['default'].omit(this._response, this._entitiesPath);

      return ___default['default'].get(responseWithoutEntities, this._metaPath, responseWithoutEntities);
    }

    get original() {
      return this._response;
    }

    get converted() {
      if (!this._isFetch) {
        return this.original;
      }

      return {
        [this._entitiesFieldName]: this.entities,
        [this._metaFieldName]: this.meta
      };
    }

  }

  const validateParticle = providedOptions => {
    if (typeof providedOptions.particle !== 'string' || providedOptions.particle === '') {
      throw new Error('"particle" must be a nonempty string');
    }
  };

  const setupOptions$1 = providedOptions => ({
    idField: 'id',
    // Response
    entityPath: 'data',
    entitiesPath: 'data',
    entitiesFieldName: 'records',
    hasMeta: true,
    metaPath: '',
    metaFieldName: 'meta',
    // Find
    findPath: (filter, {
      particle,
      idField
    }) => `/${particle}/${typeof filter === 'object' ? filter[idField] : filter}`,
    findParams: (filter, {
      idField
    }) => ___default['default'].omit(filter, idField),
    // Fetch
    fetchPath: (filter, {
      particle
    }) => `/${particle}`,
    fetchParams: (filter, {
      idField
    }) => ___default['default'].omit(filter, idField),
    // Upsert
    createMethod: 'post',
    updateMethod: 'put',
    upsertContentType: contentTypes.JSON,
    upsertPath: (data, {
      particle,
      idField
    }) => `/${particle}${data[idField] ? '/' : ''}${___default['default'].get(data, idField, '')}`,
    upsertMethod: (data, {
      idField,
      createMethod,
      updateMethod
    }) => data[idField] ? updateMethod : createMethod,
    // Remove
    removeMethod: 'remove',
    removePath: (data, {
      particle,
      idField
    } = {}) => `/${particle}/${typeof data === 'number' ? data : data[idField]}`,
    ...providedOptions
  });

  const provider = api => (providedOptions = {}) => {
    validateParticle(providedOptions);
    const options = setupOptions$1(providedOptions);

    const find = async filter => {
      const response = await api.get(options.findPath(filter, options), options.findParams(filter, options));
      return new Response(response, options);
    };

    const fetch = async filter => {
      const response = await api.get(options.fetchPath(filter, options), options.fetchParams(filter, options));
      return new Response(response, options, true);
    };

    const upsert = async (data, params) => {
      const method = options.upsertMethod(data, options);
      const response = await api[method](options.upsertPath(data, options), data, params, options.upsertContentType);
      return new Response(response, options);
    };

    const remove = async data => {
      const response = await api[options.removeMethod](options.removePath(data, options));
      return new Response(response, options);
    };

    return {
      find,
      fetch,
      upsert,
      remove
    };
  };

  exports.actions = actions;
  exports.api = api;
  exports.clearAll = clearAll;
  exports.contentTypes = contentTypes;
  exports.entity = entity;
  exports.fetch = fetch;
  exports.find = find;
  exports.provider = provider;
  exports.remove = remove;
  exports.selectAll = selectAll;
  exports.selectCollection = selectCollection;
  exports.selectMeta = selectMeta;
  exports.selectOne = selectOne;
  exports.snowboxMiddleware = snowboxMiddleware;
  exports.snowboxReducer = snowboxReducer;
  exports.statuses = statuses;
  exports.upsert = upsert;
  exports.useList = useList;

  Object.defineProperty(exports, '__esModule', { value: true });

  return exports;

}({}, axios, normalizr, react, reselect, redux, produce, _));
